"""
The RPC(Remote Procedure Call) module.

Handles GET, SET and PUT communication with the server using JSON encoded
messages.

"""
import json
import datetime
import os
import logging
from jsonrpcclient import requests, response

JSONRPC = "2.0"


class SeluxitRpc:
    """
    Check connection status.

    Invokes an init_ok() method on reference of connection. If error with
    communication occurs then a connection is closed.

    Args:
        connection: A reference to the socket instance.
        error_msg: Error message while one occurs.

    """

    @staticmethod
    def create_meta(network, network_id):
        """Create meta.

        Creates and returns a meta data about a network.

        Args:
            network:
            network_id: Identify number of a network.

        Returns:
            meta data dictionary.

        """
        return {
            "id": network_id,
            "type": "{}".format(network),
            "version": "2.0"
        }

    @staticmethod
    def get_rpc_success_response(message_id):
        """Get success response.

        Returns a success, JSON formatted, encoded in utf-8 response.

        Args:
            message_id: Id of a message.

        Returns:
            JSON formatted data.

        """
        success_response = str(response.SuccessResponse(jsonrpc=JSONRPC,
                                                        id=message_id,
                                                        result=True))
        return json.loads(success_response)

    @staticmethod
    def get_rpc_fail_response(message_id, text):
        """
        Get fail response.

        Returns a fail, JSON formatted, encoded in utf-8 response with a error
        message.

        Args:
            message_id: Id of a message.
            text: A message of error description.

        Returns:
            JSON formatted data.

        """
        error_description = {'message': text, 'code': -32020, 'data': ''}
        error_response = str(response.ErrorResponse(jsonrpc=JSONRPC,
                                                    id=message_id,
                                                    error=error_description))
        return json.loads(error_response)

    @staticmethod
    def is_upgradable():
        """
        Check if System is set to upgradable.

        Returns:
            True, if the system is upgradable,
            False, if it is not.
        """
        return True if os.getenv("UPGRADABLE") in ['true', 'True'] else False

    def __init__(self, save_init):
        """
        Initialize the seluxit_rpc class.

        Initializes an object of seluxit_rpc class by passing required
        parameters. While initialization, wapp_log is created.

        Args:
            save_init: Determines whether or not save json data.

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.save_init = save_init
        self.filename = 'Init_json.txt'
        try:
            os.remove(self.filename)
        except OSError:
            pass

    def get_rpc_network(self, network_id, network_name, put=True):
        """
        Retrieve network from server.

        Retrieves network from a server. The method is used while adding a
        network. Depending on request method it fills data_json_rpc dictionary
        with different information.

        Args:
            network_id: Unique identifying number of a network
            network_name: Name of a network
            put: defines if the request method is put {default: True}

        Returns:
            JSON formatted data of network

        """
        network = "network"
        meta = self.create_meta(network, network_id)
        if SeluxitRpc.is_upgradable():
            meta.update({'upgradable': True})
        data_inside = {
            "meta": meta,
            'name': network_name
        }

        if put:
            data_json_rpc = requests.Request('PUT',
                                             url='/{}/{}'.format(
                                                 network,
                                                 network_id),
                                             data=data_inside)
        else:
            data_json_rpc = requests.Request('POST',
                                             url='/{}'.format(network),
                                             data=data_inside)
        return data_json_rpc

    def get_rpc_state(
            self,
            data,
            network_id,
            device_id,
            value_id,
            report_id,
            set_type,
            get=False,
            put=True,
            trace_id=None,
            state_obj=None
    ):
        """
        Retrieve state of the value.

        Retrieves the value state from a server. Uses functionality of
        base_device_value to get a directory with common attributes of value
        class.

        Args:
            data: Data from the state of the value.
            network_id: Unique identifier of the network.
            device_id: Unique identifier of a device.
            value_id: Unique identifier of a value.
            report_id: Unique identifier of the report state.
            set_type: The type to set.
            get: Defines if the request is of type GET. (default: {False})
            put: Defines if the request is of type PUT. (default: {True})
            trace_id: ID of the debug trace. (default: {None})
            state_obj: Reference to the value instance's state instance.
                (default: {None})

        Returns:
            JSON formatted data of the state.

        """
        update = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        device_state = {
            'meta': {},
            'type': '',
            'status': 'Send',
            'data': data,
            'timestamp': update
        }
        state = "state"
        device_state["meta"] = self.create_meta(state, report_id)

        if state_obj is not None:
            state_obj.timestamp = update

        if set_type == 'report':
            set_type = 'Report'
        else:
            set_type = 'Control'

        device_state['type'] = set_type

        if get is True and put is False:
            put = True
        else:
            get = False

        data_json_rpc = self.create_json_message(
            device_id,
            network_id,
            value_id,
            device_state,
            put=put,
            state=state,
            state_id=report_id,
            trace_id=trace_id,
            get=get
        )
        return data_json_rpc

    def create_json_message(
            self,
            device_id,
            network_id,
            value_id,
            data,
            put,
            state=None,
            state_id=None,
            trace_id=None,
            get=False
    ):
        """
        Create a JSON encoded message.

        Creates a message that is used for communicating with a server. Both
        for put and post as well as get request methods. The message is build
        as a url and contains information about a network, device, value and
        state.

        Args:
            device_id: Unique identifying number of device.
            network_id: Unique identifying number of network.
            value_id: Unique identifying number of value.
            data: Passed data around which a message will be created.
            put: Determines whether or not it is put request.
            state: reference to a state object.
            state_id: Unique identifying number of state.
            trace_id: Id of trace is necessary.
            get: Defines if the request is of type GET. (default: {False})

        """
        base_url = '/network/{}/device/{}/value/'.format(network_id, device_id)
        if put:
            if get:
                verb = 'GET'
            else:
                verb = 'PUT'

            if state == 'state':
                url = "{}{}/{}/{}".format(base_url, value_id, state, state_id)
            else:
                url = "{}{}".format(base_url, value_id)

            if trace_id:
                url = "{}?trace={}".format(url, trace_id)
        else:
            verb = 'POST'
            if state == 'state':
                url = "{}{}/{}".format(base_url, value_id, state)
            else:
                url = base_url

        return requests.Request(verb,
                                url=url,
                                data=data)

    def get_rpc_whole_json(self, json_data):
        """
        Creates request containing the whole json file.

        The method is used while starting initializing and it fills
        data_json_rpc dictionary with url and data containing json object.

        Args:
            json_data: Data read from json file.

        Returns:
            JSON formatted data of network

        """
        data_json_rpc = requests.Request('POST',
                                         url='/{}'.format("network"),
                                         data=json_data)
        return data_json_rpc

    def add_whole_json(
            self,
            connection,
            json_data
    ):
        """Add an instance of the whole json file.

        While initializing adds network/device/value/state to send and
        receive queue.

        Args:
            connection: A reference to the socket instance.
            json_data: Data read from json file.

        """
        message = self.get_rpc_whole_json(json_data)
        self.send_init_json(connection, message)
        connection.add_id_to_confirm_list(message)

    def send_init_json(self, connection, json_data):
        """Send initial JSON data.

        Sends the initial state of the JSON data before instantiation or
        modification of it.

        Args:
            connection: Reference to the connection socket.
            json_data: Initial JSON data.

        """
        connection.create_bulk(json_data)
        if self.save_init:
            with open(self.filename, 'a+') as file:
                file.write(str(json_data) + "\n")



"""
The network module.

Stores attributes for the network instance.
"""
import logging
from .errors import wappsto_errors


class Network:
    """
    Network instance class.

    Stores attributes for the network instance.
    """

    def __init__(self, uuid, version, name):
        """
        Initialize the Network class.

        Initializes an object of network class by passing required parameters.

        Args:
            uuid: Unique identifier of a network
            version: Version of a network
            name: Name of a network

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.uuid = uuid
        self.version = version
        self.name = name
        self.callback = self.__callback_not_set
        msg = "Network {} Debug \n{}".format(name, str(self.__dict__))
        self.wapp_log.debug(msg)

    def set_callback(self, callback):
        """
        Set the callback.

        Sets the callback attribute. It will be called by the __send_logic
        method.

        Args:
            callback: Callback reference.

        Raises:
            CallbackNotCallableException: Custom exception to signify invalid
            callback.

        """
        try:
            if not callable(callback):
                msg = "Callback method should be a method"
                raise wappsto_errors.CallbackNotCallableException(msg)
            self.callback = callback
            self.wapp_log.debug("Callback {} has been set.".format(callback))
            return True
        except wappsto_errors.CallbackNotCallableException as e:
            self.wapp_log.error("Error setting callback: {}".format(e))
            raise

    def handle_delete(self):
        """
        Handle delete.

        Calls the __call_callback method with initial input of "remove".

        Returns:
            result of __call_callback method.

        """
        return self.__call_callback('remove')

    def __callback_not_set(self, network, event):
        """
        Message about no callback being set.

        Temporary method to signify that there is no callback set for the
        class.

        Returns:
            "Callback not set" message.

        """
        msg = "Callback for network '{}' is not set".format(self.name)
        return self.wapp_log.debug(msg)

    def __call_callback(self, event):
        if self.callback is not None:
            return self.callback(self, event)
        return True



"""
The object instance initialization module.

Stores the initialization functionality for the object instances of the
network, devices, values and states.
"""
import logging


class Initialize:
    """
    The initialize class handles initializing object instances.

    Using references to the RPC, socket and object instances, it sends
    activation messages to the server and enables sending/receiving for the
    network, device, value and state instances.
    """

    def __init__(self, rpc):
        """
        Create an initialization instance.

        The initialize instance contains the initialize_all method that is used
        to activate the object instances so that they can send/receive to the
        server.

        Args:
            rpc: Reference to the RPC instance.

        """
        self.rpc = rpc

    def initialize_all(self, conn, instance):
        """
        Initialize all of the devices on the sending and receiving queue.

        Adds all the devices to the send receive queue by pinging the server
        with their information in order to initialize them.

        Args:
            conn: A reference to the socket instance.
            instance: A reference to the network and device instances.

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())

        self.rpc.add_whole_json(conn, instance.build_json())

        msg = "The whole network {} added to Sending queue {}.".format(
            instance.network_cl.name,
            self.rpc
        )
        # All of the debug logs may need relocation to the RPC class.
        self.wapp_log.debug(msg)



"""
Receive and send handling module.

Handles building sending and receiving data messages and adding them to the
sending queue.
"""
import random
import logging
from . import message_data


def send_trace(sending_queue, parent, trace_id, data, control_value_id=None):
    """
    Add a trace to the sending queue.

    Adds a trace URL to the sending queue for debugging purposes.

    Args:
        sending_queue: The queue requests are being added to.
        parent: Owner of the trace URL.
        trace_id: ID of the trace message.
        data: Trace message data.
        control_value_id: ID of the control state of the value
            (default: {None})

    """
    if trace_id:
        trace = message_data.MessageData(
            message_data.SEND_TRACE,
            parent=parent,
            trace_id=trace_id,
            data=data,
            text="ok",
            control_value_id=control_value_id)
        sending_queue.put(trace)


class Handlers:
    """
    Message data handlers.

    Processes incoming and outgoing PUT, POST and GET messages.
    """

    def __init__(self, instance):
        """
        Initialize the Handler class.

        Create an instance of the Handlers class to handle incoming and
        outgoing communication with the server.

        Args:
            instance: Reference to the object instance class.

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.instance = instance

    def __get_random_id(self):
        network_n = self.instance.network_cl.name
        random_int = random.randint(1, 25000)
        return "{}{}".format(network_n, random_int)

    def handle_incoming_put(
            self,
            control_id,
            incoming_value,
            sending_queue,
            trace_id
    ):
        """
        Process an incoming PUT request.

        Handles an incoming PUT request and changes the specific value to
        the incoming value.

        Args:
            control_id: UUID of the Control state.
            incoming_value: Value received from the server.
            sending_queue: The queue requests are being added to.
            trace_id: Trace ID used to create a URL for debugging.
                (default: {None})

        Returns:
            True when successfully handling the request, False otherwise.

        """
        random_id = None
        if trace_id:
            random_id = self.__get_random_id()

        for device in self.instance.device_list:
            for value in device.value_list:
                if value.control_state is not None:
                    if value.control_state.uuid == control_id:
                        if value.handle_control(data_value=incoming_value):
                            send_trace(
                                sending_queue,
                                value.uuid,
                                trace_id,
                                incoming_value,
                                control_value_id=random_id
                            )
                            return True
                        else:
                            return False

        msg = "Unhandled put {} : {}".format(control_id, incoming_value)
        self.wapp_log.warning(msg)
        return False

    def handle_incoming_get(self, report_id, sending_queue, trace_id):
        """
        Process an incoming GET request.

        Handles an incoming GET request and sends the current value to the
        server.

        Args:
            report_id: UUID of the Report state.
            sending_queue: The queue requests are being added to.
            trace_id: Trace ID used to create a URL for debugging.
                (default: {None})


        Returns:
            True when successfully handling the request, False otherwise.

        """
        random_id = None
        if trace_id:
            random_id = self.__get_random_id()

        for device in self.instance.device_list:
            for value in device.value_list:
                if value.report_state is not None:
                    if report_id.endswith(value.report_state.uuid):
                        current_value = value.data_value
                        send_trace(
                            sending_queue,
                            value.uuid,
                            trace_id,
                            current_value,
                            control_value_id=random_id
                        )
                        # value.last_update_of_report = value.get_now()
                        value.handle_refresh()
                        return True

        self.wapp_log.warning("Unhandled get {}".format(report_id))
        return False

    def handle_incoming_delete(self, id, sending_queue, trace_id):
        """
        Handle incoming request to delete.

        Deals with requests to delete network/device/value/state, depending on
        the information provided in arguments.

        Args:
            id: ID of element to perform delete action to.
            sending_queue: Reference to queue where alements to be sent are
            saved.
            trace_id: ID used for tracing the performed actions.

        Returns:
            True or False, depending on the result received during execution.

        """
        random_id = None
        if trace_id:
            random_id = self.__get_random_id()

            send_trace(
                sending_queue,
                id,
                trace_id,
                None,
                control_value_id=random_id
            )

        if True or self.instance.network_cl.uuid == id:
            return self.instance.network_cl.handle_delete()

        for device in self.instance.device_list:
            if device.uuid == id:
                try:
                    return device.handle_delete()
                except AttributeError:
                    self.wapp_log.warning("Unhandled device delete for {}"
                                          .format(id))
                    return False
            for value in device.value_list:
                if value.uuid == id:
                    try:
                        return value.handle_delete()
                    except AttributeError:
                        self.wapp_log.warning("Unhandled value delete for {}"
                                              .format(id))
                        return False
                for state in value.state_list:
                    if state.uuid == id:
                        try:
                            return state.handle_delete()
                        except AttributeError:
                            msg = "Unhandled state delete for {}".format(id)
                            self.wapp_log.warning(msg)
                            return False

        self.wapp_log.warning("Unhandled delete {}".format(id))
        return False



#!/usr/bin/env python3
import os
import json
import pytest
import wappsto
import urllib.parse
from mock import Mock
from unittest.mock import patch

from wappsto.connection import message_data
from wappsto.object_instantiation import status
from wappsto.connection.network_classes.errors import wappsto_errors

ADDRESS = "wappsto.com"
PORT = 11006
TEST_JSON = "test_JSON/test_json.json"
TEST_JSON_prettyprint = "test_JSON/test_json_prettyprint.json"


def check_for_correct_conn(*args, **kwargs):
    if args[0][0] != ADDRESS or args[0][1] != PORT:
        raise wappsto_errors.ServerConnectionException


def fake_connect(self, address, port):
    wappsto.RETRY_LIMIT = 2
    with patch('ssl.SSLContext.wrap_socket') as context:
        context.connect = Mock(side_effect=check_for_correct_conn)
        with patch('time.sleep', return_value=None), patch('threading.Thread'), patch('wappsto.communication.ClientSocket.add_id_to_confirm_list'), patch('socket.socket'), patch('ssl.SSLContext.wrap_socket', return_value=context):
            self.service.start(address=address, port=port)


def fix_object(self, callback_exists, testing_object):
    if callback_exists:
        test_callback = Mock(return_value=True)
        testing_object.set_callback(test_callback)
    else:
        testing_object.callback = None
    return testing_object


def create_response(self, verb, callback_exists, trace_id, bulk):
    value = self.service.instance.device_list[0].value_list[0]
    value = fix_object(self, callback_exists, value)
    id = str(value.control_state.uuid)
    url = str(value.report_state.uuid)
    trace = ''

    if verb == "DELETE":
        network = self.service.get_network()
        network = fix_object(self, callback_exists, network)
        url = str(network.uuid)
    
    if verb == "DELETE" or verb == "PUT" or verb == "GET":
        if trace_id is not None:
            trace = '"meta": {"trace": "'+str(trace_id)+'"},'

        message = '{"jsonrpc": "2.0", "id": "1", "params": {"url": "'+url+'",'+trace+' "data": {"meta": {"id": "'+id+'"}, "data": "93"}}, "method": "'+verb+'"}'
    else:
        message = '{"jsonrpc": "2.0", "id": "1", "params": {"url": "/network/b03f246d-63ef-446d-be58-ef1d1e83b338/device/a0e087c1-9678-491c-ac47-5b065dea3ac0/value/7ce2afdd-3be3-4945-862e-c73a800eb209/state/a7b4f66b-2558-4559-9fcc-c60768083164", "data": {"meta": {"id": "a7b4f66b-2558-4559-9fcc-c60768083164", "type": "state", "version": "2.0"}, "type": "Report", "status": "Send", "data": "93", "timestamp": "2020-01-22T08:22:57.216500Z"}}, "method": "??????"}'

    if bulk:
        message = [message, message]
        message = str(message)

    return message


def exists_in_dictionary(key, dict):
    return True if key in dict else False

# ################################## TESTS ################################## #


class TestJsonLoadClass:
    
    @classmethod
    def setup_class(self):
        self.test_json_prettyprint_location = os.path.join(
            os.path.dirname(__file__),
            TEST_JSON_prettyprint)
        self.test_json_location = os.path.join(
            os.path.dirname(__file__),
            TEST_JSON)

    def test_load_prettyprint_json(self):
        # Arrange
        with open(self.test_json_location, "r") as json_file:
            decoded = json.load(json_file)

        # Act
        service = wappsto.Wappsto(json_file_name=self.test_json_prettyprint_location)

        # Assert
        assert service.instance.decoded == decoded


class TestConnClass:

    def setup_method(self):
        self.test_json_location = os.path.join(os.path.dirname(__file__), TEST_JSON)
        self.service = wappsto.Wappsto(json_file_name=self.test_json_location)

    @pytest.mark.parametrize("address,port,callback_exists,expected_status,value_changed_to_none,upgradable", [(ADDRESS, PORT, True, status.RUNNING, False, False),
                                                     (ADDRESS, -1, True, status.DISCONNECTING, False, False),
                                                     ("wappstoFail.com", PORT, True, status.DISCONNECTING, False, False),
                                                     (ADDRESS, PORT, False, status.RUNNING, False, False),
                                                     (ADDRESS, -1, False, status.DISCONNECTING, False, False),
                                                     ("wappstoFail.com", PORT, False, status.DISCONNECTING, False, False),
                                                     (ADDRESS, PORT, True, status.RUNNING, True, False),
                                                     (ADDRESS, -1, True, status.DISCONNECTING, True, False),
                                                     ("wappstoFail.com", PORT, True, status.DISCONNECTING, True, False),
                                                     (ADDRESS, PORT, False, status.RUNNING, True, False),
                                                     (ADDRESS, -1, False, status.DISCONNECTING, True, False),
                                                     ("wappstoFail.com", PORT, False, status.DISCONNECTING, True, False),
                                                     (ADDRESS, PORT, True, status.RUNNING, False, True),
                                                     (ADDRESS, -1, True, status.DISCONNECTING, False, True),
                                                     ("wappstoFail.com", PORT, True, status.DISCONNECTING, False, True),
                                                     (ADDRESS, PORT, False, status.RUNNING, False, True),
                                                     (ADDRESS, -1, False, status.DISCONNECTING, False, True),
                                                     ("wappstoFail.com", PORT, False, status.DISCONNECTING, False, True),
                                                     (ADDRESS, PORT, True, status.RUNNING, True, True),
                                                     (ADDRESS, -1, True, status.DISCONNECTING, True, True),
                                                     ("wappstoFail.com", PORT, True, status.DISCONNECTING, True, True),
                                                     (ADDRESS, PORT, False, status.RUNNING, True, True),
                                                     (ADDRESS, -1, False, status.DISCONNECTING, True, True),
                                                     ("wappstoFail.com", PORT, False, status.DISCONNECTING, True, True)])
    def test_connection(self, address, port, callback_exists, expected_status, value_changed_to_none, upgradable):
        # Arrange
        status_service = self.service.get_status()
        fix_object(self, callback_exists, status_service)
        if value_changed_to_none:
            self.service.instance.network_cl.name = None

        # Act
        with patch('os.getenv', return_value=str(upgradable)):
            try:
                fake_connect(self, address, port)
                args, kwargs = self.service.socket.my_socket.send.call_args
                arg = json.loads(args[0].decode('utf-8'))
                sent_json = arg[0]['params']['data']
            except wappsto_errors.ServerConnectionException:
                sent_json = None
                pass

        # Assert
        if sent_json != None:
            assert not 'None' in str(sent_json)
            assert (upgradable and 'upgradable' in str(sent_json['meta']) or
                    not upgradable and not 'upgradable' in str(sent_json['meta']))
        assert self.service.status.get_status() == expected_status

class TestValueSendClass:

    @classmethod
    def setup_class(self):
        test_json_location = os.path.join(os.path.dirname(__file__), TEST_JSON)
        self.service = wappsto.Wappsto(json_file_name=test_json_location)
        fake_connect(self, ADDRESS, PORT)

    @pytest.mark.parametrize("input,step_size,expected", [(8, 1, "8"),# value on the step
                                                     (8, -1, "8"),
                                                     (-8, 1, "-8"),
                                                     (-8, -1, "-8"),
                                                     (100, 1, "100"),
                                                     (-100, 1, "-100"),
                                                     (0, 1, "0"),
                                                     (-0, 1, "0"),
                                                     (-99.9, 1, "-100"),# decimal value
                                                     (-0.1, 1, "-1"),
                                                     (0.1, 1, "0"),
                                                     (3.3, 1, "3"),
                                                     (3.0, 1, "3"),
                                                     (3.9, 1, "3"),
                                                     (-0.1, 1, "-1"),
                                                     (-3.3, 1, "-4"),
                                                     (-3.0, 1, "-3"),
                                                     (-3.9, 1, "-4"),
                                                     (-101, 1, None),# out of range
                                                     (101, 1, None),
                                                     (3, 2, "2"),# big steps
                                                     (3.999, 2, "2"),
                                                     (4, 2, "4"),
                                                     (-3, 2, "-4"),
                                                     (-3.999, 2, "-4"),
                                                     (-4, 2, "-4"),
                                                     (1, 0.5, "1"),# decimal steps
                                                     (1.01, 0.02, "1"),
                                                     (2.002, 0.02, "2"),
                                                     (2.002, 0.0002, "2.002"),
                                                     (-1, 0.5, "-1"),
                                                     (-1.01, 0.02, "-1.02"),
                                                     (-2.002, 0.02, "-2.02"),
                                                     (-2.002, 0.0002, "-2.002"),
                                                     (2, 1.0e-07, "2"),
                                                     (2, 123.456e-5, "1.9999872"),
                                                     (1, 9.0e-20, "0.99999999999999999999")])
    def test_send_value_update(self, input, step_size, expected):
        # Arrange
        self.service.socket.message_received = True
        self.service.socket.my_socket.send = Mock()
        device = self.service.get_devices()[0]
        value = device.value_list[0]
        value.number_step = step_size

        # Act
        try:
            value.update(input)
            args, kwargs = self.service.socket.my_socket.send.call_args
            arg = json.loads(args[0].decode('utf-8'))
            result = arg[0]['params']['data']['data']
        except TypeError:
            result = None

        # Assert
        assert result == expected

    @classmethod
    def teardown_class(self):
        self.service.stop()


class TestReceiveThreadClass:

    @classmethod
    def setup_class(self):
        test_json_location = os.path.join(os.path.dirname(__file__), TEST_JSON)
        self.service = wappsto.Wappsto(json_file_name=test_json_location)
        fake_connect(self, ADDRESS, PORT)

    @pytest.mark.parametrize("id,verb,callback_exists,trace_id,expected_rpc_id,expected_msg_id,expected_trace_id",
                             [(1, 'PUT', True, None, '1', message_data.SEND_SUCCESS, None),
                              (1, 'PUT', False, None, '1', message_data.SEND_FAILED, None),
                              (1, 'DELETE', True, None, '1', message_data.SEND_SUCCESS, None),
                              (1, 'DELETE', False, None, '1', message_data.SEND_SUCCESS, None),
                              (1, 'GET', True, None, '1', message_data.SEND_SUCCESS, None),
                              (1, 'GET', False, None, '1', message_data.SEND_SUCCESS, None),
                              (1, 'wrong_verb', False, None, '1', message_data.SEND_FAILED, None),
                              (1, 'wrong_verb', True, None, '1', message_data.SEND_FAILED, None),
                              (1, 'PUT', True, 321, None, message_data.SEND_TRACE, '321'),
                              (1, 'PUT', False, 321, '1', message_data.SEND_FAILED, None),
                              (1, 'DELETE', True, 321, None, message_data.SEND_TRACE, '321'),
                              (1, 'DELETE', False, 321, None, message_data.SEND_TRACE, '321'),
                              (1, 'GET', True, 321, None, message_data.SEND_TRACE, '321'),
                              (1, 'GET', False, 321, None, message_data.SEND_TRACE, '321'),
                              (1, 'wrong_verb', False, 321, '1', message_data.SEND_FAILED, None),
                              (1, 'wrong_verb', True, 321, '1', message_data.SEND_FAILED, None)
                             ])
    @pytest.mark.parametrize("bulk", [False, True])
    def test_receive_thread_method(self, id, verb, callback_exists, trace_id, 
                                   expected_rpc_id, expected_msg_id, expected_trace_id, bulk):
        # Arrange
        response = create_response(self, verb, callback_exists, trace_id, bulk)
        self.service.socket.my_socket.recv = Mock(side_effect=[response.encode('utf-8'), KeyboardInterrupt])

        # Act
        try:
            # runs until mock object is run and its side_effect raises
            # exception
            self.service.socket.receive_thread()
        except KeyboardInterrupt:
            pass

        # Assert
        while self.service.socket.sending_queue.qsize() > 0:
            send = self.service.socket.sending_queue.get()
            assert (send.msg_id == message_data.SEND_SUCCESS or 
                    send.msg_id == expected_msg_id)

    @pytest.mark.parametrize("id,type", [(93043873, "error"),(93043873, "result")])
    def test_receive_thread_other(self, id, type):
        # Arrange
        response = '{"jsonrpc": "2.0", "id": "'+ str(id) +'", "'+type+'": {"value": "True", "meta": {"server_send_time": "2020-01-22T08:22:55.315Z"}}}'
        self.service.socket.packet_awaiting_confirm[str(id)] = response
        self.service.socket.my_socket.recv = Mock(side_effect=[response.encode('utf-8'), KeyboardInterrupt])

        # Act
        try:
            # runs until mock object is run and its side_effect raises
            # exception
            self.service.socket.receive_thread()
        except KeyboardInterrupt:
            pass

        # Assert
        assert len(self.service.socket.packet_awaiting_confirm) == 0

    @classmethod
    def teardown_class(self):
        self.service.stop()


class TestSendThreadClass:

    def setup_method(self):
        test_json_location = os.path.join(os.path.dirname(__file__), TEST_JSON)
        self.service = wappsto.Wappsto(json_file_name=test_json_location)
        fake_connect(self, ADDRESS, PORT)

    @pytest.mark.parametrize("type", [message_data.SEND_SUCCESS,
                                      message_data.SEND_REPORT,
                                      message_data.SEND_FAILED,
                                      message_data.SEND_RECONNECT,
                                      message_data.SEND_CONTROL])
    @pytest.mark.parametrize("value,expected_value", [('test_value','test_value'),
                                                                ('', None),
                                                                (None, None),
                                                                ([],None)])
    @pytest.mark.parametrize("messages_in_queue", [1, 2])
    def test_send_thread(self, type, messages_in_queue, value, expected_value):
        # Arrange
        self.service.socket.message_received = True
        self.service.get_network().name = value
        i = 0
        while i < messages_in_queue:
            i += 1
            reply = message_data.MessageData(
                type,
                rpc_id=value,
                data=value
            )
            self.service.socket.sending_queue.put(reply)
        self.service.socket.my_socket.send = Mock(side_effect=KeyboardInterrupt)
        self.service.socket.add_id_to_confirm_list = Mock()

        # Act
        try:
            # runs until mock object is run and its side_effect raises
            # exception
            self.service.socket.send_thread()
        except KeyboardInterrupt:
            args, kwargs = self.service.socket.my_socket.send.call_args
            arg = args[0].decode('utf-8')
            requests = json.loads(arg)

        # Assert
        assert self.service.socket.sending_queue.qsize() == 0
        assert messages_in_queue == len(requests)
        for request in requests:
            if type == message_data.SEND_SUCCESS:
                assert request.get('id', None) == expected_value
                assert bool(request['result']) == True
            elif type == message_data.SEND_FAILED:
                assert request.get('id', None) == expected_value
                assert request['error'] == {"code": -32020}
            elif type == message_data.SEND_REPORT:
                assert request['params']['data'].get('data', None) == expected_value
                assert request['params']['data']['type'] == "Report"
                assert request['method'] == "PUT"
            elif type == message_data.SEND_RECONNECT:
                assert request['params']['data'].get('name', None) == expected_value
                assert request['params']['data']['meta']['type'] == "network"
                assert request['method'] == "POST"
            elif type == message_data.SEND_CONTROL:
                assert request['params']['data'].get('data', None) == expected_value
                assert request['params']['data']['type'] == "Control"
                assert request['method'] == "PUT"

    @pytest.mark.parametrize("rpc_id,expected_trace_id,type", [(93043873, 332, message_data.SEND_TRACE)])
    def test_send_thread_send_trace(self, rpc_id, expected_trace_id, type):
        # Arrange
        reply = message_data.MessageData(
            type,
            trace_id = expected_trace_id,
            rpc_id=rpc_id
        )
        self.service.socket.sending_queue.put(reply)

        # Act
        with patch('urllib.request.urlopen', side_effect=KeyboardInterrupt) as urlopen:
            try:
                # runs until mock object is run and its side_effect raises
                # exception
                self.service.socket.send_thread()
            except KeyboardInterrupt:
                args, kwargs = urlopen.call_args
                arg = urllib.parse.parse_qs(args[0])
        result_trace_id = int(arg['https://tracer.iot.seluxit.com/trace?id'][0])

        # Assert
        assert result_trace_id == expected_trace_id
